/*******
* Implementation of M-Tree (Authors: Paolo Ciaccia, Marco Patella, Pavel Zezula)
* M-tree Web Page: http://www-db.deis.unibo.it/Mtree/
*************************************************
*
* Copyright (C) 2009 Ricardo J. Barrientos Rojel (ribarrie@dcc.uchile.cl)
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License (http://www.gnu.org/licenses/gpl.txt)
* as published by the Free Software Foundation; either version 2 of the
* License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
**********/

#include "estruc.h"

extern long int cuentaeditdist;

void insert(Node *N, TipoUser dato, TipoDist dparent);
void Split(Node *N, Ientry *datoI, Lentry *datoL);
void up_dp_N(Node *N, Ientry *ie);//update the distance to the parent of all the entries of N. ie is the parent entry of N.
void up_dp(Ientry *ie, Node *N);//update the distance to the parent
void up_rc(Ientry *Op, int T);
//partition implements the strategy "Generalized Hyperplane"
int *partition(Ientry *arr_i, Lentry *arr_l, int quan, Ientry *Op1, Ientry *Op2, Ientry *N1I, Ientry *N2I, Lentry *N1L, Lentry *N2L);
//promote implements the strategy mM_RAD
void promote(Ientry *arr_i, Lentry *arr_l, int quan, Ientry *Op1, Ientry *Op2);
TipoDist getrc(Ientry *arr_i, Lentry *arr_l, int quan, int ind);//get the rc of arr[ind] for the set arr (arr_i or arr_l)
void Save_ex(Node *N);
void Save();

Node *MTree=NULL;
FILE *pf_sal;

main(int argc, char *argv[])
{
	int i,j;
	FILE *pf;
	TipoUser dato;

	if(argc < 2)
	{
		printf("Missing Arguments :: Run : a.out file_BD\n");
		return 0;
	}

	pf = fopen(argv[1], "r");

	//Creation of the MTree
	MTree = newNode(LEAF_ROOT);

	//Insertion of the elements
	while (leedato(dato, pf)!=ERROR && !feof(pf))
	{
		printf("Inserting : %s...", dato);
		fflush(stdout);
		insert(MTree, dato, 0);
		printf(" OK\n");
		fflush(stdout);
	}
//	imprime(MTree);
	Save();//save to disk
	printf("\nEND.\n");
	fflush(stdout);
	return 0;
}

void Save()
{
	int i;

	pf_sal = fopen("mtree.dat", "w");

	fwrite(MTree, sizeof(Node), 1, pf_sal);
	if (MTree->type == ROOT)
	{
		for(i=0; i<MTree->nro_entries; i++)
			Save_ex(MTree->data.intern[i].child);
	}
	else
	{
		printf(" [MTree->type=LEAF_ROOT] ");
	}
	fclose(pf_sal);
	return;
}

void Save_ex(Node *N)
{
	int i;

	fwrite(N, sizeof(Node), 1, pf_sal);
	if (N->type == INTERN)
	{
		for (i=0; i<N->nro_entries; i++)
			Save_ex(N->data.intern[i].child);
	}
	return;
}

void insert(Node *N, TipoUser dato, TipoDist dparent)
{
	Lentry *ltemp;
	TipoDist dist[MAX_ENTRY], min, t;
	int itemp[MAX_ENTRY], cont=0, i=0, imin;

	if (N->type != LEAF && N->type != LEAF_ROOT)//INTERN or ROOT
	{
		for (i=0; i<N->nro_entries; i++)
		{
			dist[i] = editdist(N->data.intern[i].info, dato);
			if (dist[i] <= N->data.intern[i].rc)
			{
				itemp[cont] = i;
				cont++;
			}
		}

		if (cont != 0)
		{
			min = dist[itemp[0]];
			imin = itemp[0];
			for (i=1; i<cont; i++)
			{
				if (dist[itemp[i]] < min)
				{
					min = dist[itemp[i]];
					imin = itemp[i];
				}
			}
		}
		else
		{
			imin = 0;
			min = dist[0] - N->data.intern[0].rc;
			for (i=1; i<N->nro_entries; i++)
			{
				if ((t=dist[i] - N->data.intern[i].rc) < min)
				{
					min = t;
					imin = i;
				}
			}
			N->data.intern[imin].rc = dist[imin];
		}
		insert(N->data.intern[imin].child, dato, dist[imin]);
	}
	else //Leaf
	{
		if (N->nro_entries < MAX_ENTRY)//if isn't full
		{
			copiavalor(N->data.leaf[N->nro_entries].info, dato);
			N->data.leaf[N->nro_entries].oid = ID;
			ID++;
			N->data.leaf[N->nro_entries].Dparent = dparent;
			N->nro_entries += 1;
		}
		else
		{
			ltemp = (Lentry *)malloc(sizeof(Lentry));
			copiavalor(ltemp->info, dato);
			ltemp->oid = ID;
			ID++;
			ltemp->Dparent = dparent;
			Split(N, NULL, ltemp);
		}
	}
}

void Split(Node *N, Ientry *datoI, Lentry *datoL)
{
	Ientry *eparent=NULL, Op1, Op2;
	int i=0;
	Node *newN=NULL, *newroot=NULL, *Nparent=NULL;
	Ientry arr_i[MAX_ENTRY+1], N1I[MAX_ENTRY], N2I[MAX_ENTRY];
	Lentry arr_l[MAX_ENTRY+1], N1L[MAX_ENTRY], N2L[MAX_ENTRY];
	int *lim, quan=0;

	if (N->type == LEAF || N->type == LEAF_ROOT)
	{
		if (datoL == NULL)
		{
			printf("\nError in Split :: datoL is NULL\n");
			exit(0);
		}
		for (i=0; i<N->nro_entries; i++)
		{
			copyleaf(&(arr_l[i]), &(N->data.leaf[i]));
			quan++;
		}
		copyleaf(&(arr_l[i]), datoL);
		quan++;
	}
	else
	{
		if (datoI == NULL)
		{
			printf("\nError in Split :: datoI is NULL\n");
			exit(0);
		}
		for (i=0; i<N->nro_entries; i++)
		{
			copyintern(&(arr_i[i]), &(N->data.intern[i]));
			quan++;
		}
		copyintern(&(arr_i[i]), datoI);
		quan++;
	}

	if (N->type != ROOT && N->type != LEAF_ROOT)
	{
		eparent = N->E_parent;
		Nparent = N->N_parent;
	}

	//Allocate
	newN = newNode(LEAF);

	//Promote
	if (N->type == LEAF || N->type == LEAF_ROOT)
		promote(NULL, arr_l, quan, &Op1, &Op2);
	else
		promote(arr_i, NULL, quan, &Op1, &Op2);

	//Partition
	if (N->type == LEAF || N->type == LEAF_ROOT)
		lim = partition(NULL, arr_l, quan, &Op1, &Op2, NULL, NULL, N1L, N2L);//note:there is to link from Op at his child
	else
		lim = partition(arr_i, NULL, quan, &Op1, &Op2, N1I, N2I, NULL, NULL);

	//Store N1's entries in N and N2's entries in newN
	if (N->type == LEAF || N->type == LEAF_ROOT)
	{
		for(i=0; i<lim[0]; i++)
			copyleaf(&(N->data.leaf[i]), &(N1L[i]));
		N->nro_entries = lim[0];

		newN->type = LEAF;
		for(i=0; i<lim[1]; i++)
			copyleaf(&(newN->data.leaf[i]), &(N2L[i]));
		newN->nro_entries = lim[1];
	}
	else
	{
		for(i=0; i<lim[0]; i++)
		{
			copyintern(&(N->data.intern[i]), &(N1I[i]));
			//update the parent of the child (because Op2)
			N->data.intern[i].child->N_parent = N;
			N->data.intern[i].child->E_parent = &(N->data.intern[i]);
		}
		N->nro_entries = lim[0];

		newN->type = INTERN;
		for(i=0; i<lim[1]; i++)
		{
			copyintern(&(newN->data.intern[i]), &(N2I[i]));
			//update the parent of the child (because Op2)
			newN->data.intern[i].child->N_parent = newN;
			newN->data.intern[i].child->E_parent = &(newN->data.intern[i]);
		}
		newN->nro_entries = lim[1];
	}
	free(lim);
	//Set the correct child and Dparents
	Op1.child = N;
	Op2.child = newN;
	up_dp_N(N, &Op1);
	up_dp_N(newN, &Op2);
	//Update rc
	if (N->type == LEAF || N->type == LEAF_ROOT)
	{
		up_rc(&Op1, LEAF);
		up_rc(&Op2, LEAF);
	}
	else
	{
		up_rc(&Op1, INTERN);
		up_rc(&Op2, INTERN);
	}

	if (N->type == ROOT || N->type == LEAF_ROOT)//if N is the current root
	{
		newroot = newNode(ROOT);
		copyintern(&(newroot->data.intern[0]), &Op1);
		newroot->data.intern[0].Dparent = (TipoDist)0;

		copyintern(&(newroot->data.intern[1]), &Op2);
		newroot->data.intern[1].Dparent = (TipoDist)0;
		newroot->nro_entries = 2;
		//Set the correct parent
		N->E_parent = &(newroot->data.intern[0]);
		N->N_parent = newroot;
		newN->E_parent = &(newroot->data.intern[1]);
		newN->N_parent = newroot;
		//Set the type
		if (N->type == LEAF_ROOT)
			N->type = LEAF;
		else
			N->type = INTERN;
		MTree = newroot;
	}
	else
	{
		copyintern(eparent, &Op1);
		up_dp(eparent, Nparent);//update Distance to parent
		N->E_parent = eparent;//update the parent
		N->N_parent = Nparent;
		if (N->N_parent->nro_entries == MAX_ENTRY)//if the parent is full
			Split(N->N_parent, &Op2, NULL);
		else //Store Op2 in the parent
		{
			copyintern(&(N->N_parent->data.intern[N->N_parent->nro_entries]), &Op2);
			up_dp(&(N->N_parent->data.intern[N->N_parent->nro_entries]), N->N_parent);//update Distance to parent
			newN->E_parent = &(N->N_parent->data.intern[N->N_parent->nro_entries]);//update the parent
			newN->N_parent = Nparent;
			N->N_parent->nro_entries += 1;
		}
	}
	return;
}

void up_dp(Ientry *ie, Node *N)//update the distance to the parent
{
	if (N->type==ROOT || N->type==LEAF_ROOT)
	{
		ie->Dparent = (TipoDist)0;
		return;
	}
	ie->Dparent = editdist(ie->info, N->E_parent->info);
	return;
}

void up_dp_N(Node *N, Ientry *ie)//update the distance to the parent of all the entries of N. ie is the parent entry of N.
{
	int i;
	if (N->type == LEAF || N->type == LEAF_ROOT)
	{
		for (i=0; i<N->nro_entries; i++)
			N->data.leaf[i].Dparent = editdist(N->data.leaf[i].info, ie->info);
	}
	else
	{
		for (i=0; i<N->nro_entries; i++)
			N->data.intern[i].Dparent = editdist(N->data.intern[i].info, ie->info);
	}
	return;
}

void up_rc(Ientry *Op, int T)
{
	int i;
	Node *N;
	TipoDist d, rc;

	N = Op->child;
	if (T == LEAF)
	{
		rc = editdist(Op->info, N->data.leaf[0].info);
		for (i=1; i < N->nro_entries; i++)
		{
			if ((d = editdist(Op->info, N->data.leaf[i].info)) > rc)
				rc = d;
		}
		Op->rc = rc;
	}
	else
	{
		rc = editdist(Op->info, N->data.intern[0].info) + N->data.intern[0].rc;
		for (i=1; i < N->nro_entries; i++)
		{
			d = editdist(Op->info, N->data.intern[i].info);
			if (d + N->data.intern[i].rc > rc)
				rc = d + N->data.intern[i].rc;
		}
		Op->rc = rc;

	}
	return;
}


//Implements the strategy "Generalized Hyperplane"
int *partition(Ientry *arr_i, Lentry *arr_l, int quan, Ientry *Op1, Ientry *Op2, Ientry *N1I, Ientry *N2I, Lentry *N1L, Lentry *N2L)
{
	int i, ind_n1=0, ind_n2=0, *lim;
	TipoDist d1, d2;

	lim = (int *)malloc(sizeof(int)*2);
	if (arr_i == NULL)//array of leafs
	{
		for (i=0; i<quan; i++)
		{
			if ((d1=editdist(arr_l[i].info, Op1->info)) < (d2=editdist(arr_l[i].info, Op2->info)))
			{
				copiavalor(N1L[ind_n1].info, arr_l[i].info);
				N1L[ind_n1].Dparent = arr_l[i].Dparent;
				N1L[ind_n1].oid = arr_l[i].oid;
				ind_n1++;
			}
			else
			{
				copiavalor(N2L[ind_n2].info, arr_l[i].info);
				N2L[ind_n2].Dparent = arr_l[i].Dparent;
				N2L[ind_n2].oid = arr_l[i].oid;
				ind_n2++;
			}
		}
		lim[0] = ind_n1;
		lim[1] = ind_n2;
	}
	else//array of interns
	{
		for (i=0; i<quan; i++)
		{
			if ((d1=editdist(arr_i[i].info, Op1->info)) < (d2=editdist(arr_i[i].info, Op2->info)))
			{
				copiavalor(N1I[ind_n1].info, arr_i[i].info);
				N1I[ind_n1].Dparent = arr_i[i].Dparent;
				N1I[ind_n1].rc = arr_i[i].rc;
				N1I[ind_n1].child = arr_i[i].child;
				ind_n1++;
			}
			else
			{
				copiavalor(N2I[ind_n2].info, arr_i[i].info);
				N2I[ind_n2].Dparent = arr_i[i].Dparent;
				N2I[ind_n2].rc = arr_i[i].rc;
				N2I[ind_n2].child = arr_i[i].child;
				ind_n2++;
			}
		}
		lim[0] = ind_n1;
		lim[1] = ind_n2;
	}

	return lim;
}

//promote implements the strategy mM_RAD
void promote(Ientry *arr_i, Lentry *arr_l, int quan, Ientry *Op1, Ientry *Op2)
{
	int i, j, ind_i, ind_j;
	TipoDist min, arr_rc[MAX_ENTRY+1];

	if (arr_i == NULL)
	{
		for (i=0; i<quan; i++)
				arr_rc[i] = getrc(NULL, arr_l, quan, i);

		for (i=0; i<quan-1; i++)
			for (j=i+1; j<quan; j++)
			{
				if (arr_rc[i] < arr_rc[j])
				{
					if (i==0 && j==1)
					{
						min = arr_rc[j];
						ind_i = i;
						ind_j = j;
					}
					if (arr_rc[j] < min)
					{
						min = arr_rc[j];
						ind_i = i;
						ind_j = j;
					}
				}
				else
				{
					if (i==0 && j==1)
					{
						min = arr_rc[i];
						ind_i = i;
						ind_j = j;
					}
					if (arr_rc[i] < min)
					{
						min = arr_rc[i];
						ind_i = i;
						ind_j = j;
					}
				}
			}

		copiavalor(Op1->info, arr_l[ind_i].info);
		Op1->rc = arr_rc[ind_i];

		copiavalor(Op2->info, arr_l[ind_j].info);
		Op2->rc = arr_rc[ind_j];
	}
	else
	{
		for (i=0; i<quan; i++)
				arr_rc[i] = getrc(arr_i, NULL, quan, i);

		for (i=0; i<quan; i++)
			for (j=i+1; j<quan; j++)
			{
				if (arr_rc[i] < arr_rc[j])
				{
					if (i==0 && j==1)
					{
						min = arr_rc[j];
						ind_i = i;
						ind_j = j;
					}
					if (arr_rc[j] < min)
					{
						min = arr_rc[j];
						ind_i = i;
						ind_j = j;
					}
				}
				else
				{
					if (i==0 && j==1)
					{
						min = arr_rc[i];
						ind_i = i;
						ind_j = j;
					}
					if (arr_rc[i] < min)
					{
						min = arr_rc[i];
						ind_i = i;
						ind_j = j;
					}
				}
			}

		copiavalor(Op1->info, arr_i[ind_i].info);
		Op1->rc = arr_rc[ind_i];

		copiavalor(Op2->info, arr_i[ind_j].info);
		Op2->rc = arr_rc[ind_j];
	}
	return;
}


TipoDist getrc(Ientry *arr_i, Lentry *arr_l, int quan, int ind)//get the rc of arr[ind] for the set arr (arr_i or arr_l)
{
	int i;
	TipoDist d, max=0;

	if (arr_i == NULL)//array of leaf entries
	{
		for (i=0; i<quan; i++)
		{
			if (ind != i)
				if ((d=editdist(arr_l[ind].info, arr_l[i].info)) > max)
					max = d;
		}
	}
	else//array of intern entries
	{
		for (i=0; i<quan; i++)
		{
			if (ind != i)
			{
				d = editdist(arr_i[ind].info, arr_i[i].info);
				if ((d + arr_i[i].rc) > max)
					max = d + arr_i[i].rc;
			}
		}
	}

	return max;
}





