/*******
* Implementation of M-Tree (Authors: Paolo Ciaccia, Marco Patella, Pavel Zezula)
* M-tree Web Page: http://www-db.deis.unibo.it/Mtree/
*************************************************
*
* Copyright (C) 2009 Ricardo J. Barrientos Rojel (ribarrie@dcc.uchile.cl)
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License (http://www.gnu.org/licenses/gpl.txt)
* as published by the Free Software Foundation; either version 2 of the
* License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
**********/

#include "estruc.h"
//for time
#include <sys/resource.h>
#include <time.h>
#include <sys/time.h>

#define NQ 2 //NQ = Num. of queries
#define RADIUS 1 

extern long int cuentaeditdist;
long int results=0;
FILE *pf;

void RS(Node *N, TipoUser dato, TipoDist r, TipoDist par_q);
Node *Load();
void Load_ex(Node **N, Ientry *eparent, Node *Nparent);


main(int argc, char *argv[])
{
	int i,j, cont=0;
	FILE *pf_q;
	TipoUser dato;
	Node *MTree=NULL;
	struct rusage r1, r2;
	float user_time, sys_time, real_time;
	struct timeval t1, t2;

	if(argc < 2)
	{
		printf("Missing Arguments :: Run : a.out file_queries\n");
		return 0;
	}

	pf_q = fopen(argv[1], "r");

	//Loading MTree
	printf("\nLoading MTree... ");
	fflush(stdout);
	pf = fopen("mtree.dat", "r");
	MTree = Load();
	if (MTree == NULL)
		return 0;
	printf("OK\n");

    getrusage(RUSAGE_SELF, &r1);
	gettimeofday(&t1, 0);
	cont=0;
	while (leedato(dato, pf_q)!=ERROR && !feof(pf))
	{
		RS(MTree, dato, RADIUS, 0);
		cont++;
	}
	gettimeofday(&t2, 0);
  	getrusage(RUSAGE_SELF, &r2);
	user_time = (r2.ru_utime.tv_sec - r1.ru_utime.tv_sec) + (r2.ru_utime.tv_usec - r1.ru_utime.tv_usec)/1000000.0;
	sys_time = (r2.ru_stime.tv_sec - r1.ru_stime.tv_sec) + (r2.ru_stime.tv_usec - r1.ru_stime.tv_usec)/1000000.0;
	real_time = (t2.tv_sec - t1.tv_sec) + (float)(t2.tv_usec - t1.tv_usec)/1000000;

	printf("\nRadius = %d\n", RADIUS);
	printf("E.D. = %ld\n", cuentaeditdist);
	printf("E.D. average per query = %d\n", (int)(cuentaeditdist/cont));
	printf("CPU Time = %f\n", sys_time + user_time);
	printf("Real Time = %f\n", real_time);
	fclose(pf_q);
	fclose(pf);
	return 0;
}

Node *Load()
{
	int i;
	Node *N;

	N = (Node *)malloc(sizeof(Node));
	fread(N, sizeof(Node), 1, pf);
	if (N->type != LEAF_ROOT && N->type != ROOT)
	{
		printf("\nError : type of the root different of ROOT or LEAF_ROOT\n");
		return NULL;
	}
	N->E_parent = NULL;
	N->N_parent = NULL;
	if (N->type == ROOT)
	{
		for(i=0; i<N->nro_entries; i++)
			Load_ex(&(N->data.intern[i].child), &(N->data.intern[i]), N);
	}
	else
	{
		printf("[type=LEAF_ROOT] ");
	}
	return N;
}

void Load_ex(Node **N, Ientry *eparent, Node *Nparent)
{
	int i;

	*N = (Node *)malloc(sizeof(Node));
	fread(*N, sizeof(Node), 1, pf);
	(*N)->E_parent = eparent;
	(*N)->N_parent = Nparent;
	if ((*N)->type == INTERN)
	{
		for (i=0; i<(*N)->nro_entries; i++)
			Load_ex(&((*N)->data.intern[i].child), &((*N)->data.intern[i]), *N);
	}
	return;
}


//Range Search
void RS(Node *N, TipoUser dato, TipoDist r, TipoDist par_q)
{
	int i;
	TipoDist d;

	if (N->type!=LEAF && N->type!=LEAF_ROOT)//if N is not a leaf
	{
		for (i=0; i<N->nro_entries; i++)
		{
			if (v_abs(par_q - N->data.intern[i].Dparent) <= r + N->data.intern[i].rc)
				if ((d=editdist(N->data.intern[i].info, dato)) <= r + N->data.intern[i].rc)
					RS(N->data.intern[i].child, dato, RADIUS, d);
		}
	}
	else
	{
		for (i=0; i<N->nro_entries; i++)
		{
			if (v_abs(par_q - N->data.leaf[i].Dparent) <= r)
				if ((d=editdist(N->data.leaf[i].info, dato)) <= r)
				{
//					printf("%s\n", N->data.leaf[i].info);	
					results++;
				}
		}
	}
	return;
}




