/*******
* Implementation of M-Tree
* Copyright (C) 2009 Ricardo J. Barrientos Rojel (ribarrie@dcc.uchile.cl)
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License (http://www.gnu.org/licenses/gpl.txt)
* as published by the Free Software Foundation; either version 2 of the
* License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
**********/

#include "datas.h"

int ID = 0;

struct _node;
typedef struct _node Node;

struct _ientry //Intern Object entry
{
	TipoUser info;//data
	TipoDist rc;//covering radius
	TipoDist Dparent;//distance to the parent
	Node *child;
};
typedef struct _ientry Ientry;

struct _lentry //Leaf Object entry
{
	TipoUser info;//data
	TipoDist Dparent;//distance to the parent
	int oid;//object identifier
};
typedef struct _lentry Lentry;

union u
{
	Ientry intern[MAX_ENTRY];
	Lentry leaf[MAX_ENTRY];
};
typedef union u U;

struct _node
{
	char type; //type : INTERN, LEAF, ROOT or LEAF-ROOT
	int nro_entries;
	U data;
	Ientry *E_parent;//Entry parent
	Node *N_parent;//Node parent
};

//Functions
void visit(Node *N);
void imprime(Node *N);
void copyleaf(Lentry *a, Lentry *b);
void copyintern(Ientry *a, Ientry *b);
Node *newNode(int t);


Node *newNode(int t)
{
	Node *n;
	n = (Node *)malloc(sizeof(Node));
	n->type = t;
	n->nro_entries = 0;
	n->E_parent = NULL;
	n->N_parent = NULL;
	return n;
}

void copyintern(Ientry *a, Ientry *b)
{
	copiavalor(a->info, b->info);
	a->rc = b->rc;
	a->Dparent = b->Dparent;
	a->child = b->child;
}

void copyleaf(Lentry *a, Lentry *b)
{
	copiavalor(a->info, b->info);
	a->oid = b->oid;
	a->Dparent = b->Dparent;
}


void imprime(Node *N)
{
	int i;

	printf("Node = %p\n", N);
	if (N->type == ROOT)
		printf("Type = ROOT\n");
	if (N->type == LEAF_ROOT)
		printf("Type = LEAF_ROOT\n");
	if (N->type == INTERN)
		printf("Type = INTERN\n");
	if (N->type == LEAF)
		printf("Type = LEAF\n");
	printf("nro_entries = %d\n", N->nro_entries);
	printf("E_parent = %p\n", N->E_parent);
	printf("N_parent = %p\n", N->N_parent);
	printf("data : \n");
	if (N->type == LEAF || N->type == LEAF_ROOT)
	{
		for (i=0; i<N->nro_entries; i++)
		{
			printf("\tinfo: [%s]\n", N->data.leaf[i].info);
			printf("\tDparent: %d\n", N->data.leaf[i].Dparent);
			printf("\toid: %d\n\n", N->data.leaf[i].oid);
		}
	}
	else
	{
		for (i=0; i<N->nro_entries; i++)
		{
			printf("\tIentry: [%p]\n", &(N->data.intern[i]));
			printf("\tinfo: [%s]\n", N->data.intern[i].info);
			printf("\trc: %d\n", N->data.intern[i].rc);
			printf("\tDparent: %d\n", N->data.intern[i].Dparent);
			printf("\tchild: %p\n\n", N->data.intern[i].child);
		}
		for (i=0; i<N->nro_entries; i++)
			visit(N->data.intern[i].child);
	}
	printf("--------------------------------\n\n");

	return;
}

void visit(Node *N)
{
	int i;

	printf("Dir. Nodo = %p\n", N);
	if (N->type == ROOT)
		printf("Type = ROOT\n");
	if (N->type == LEAF_ROOT)
		printf("Type = LEAF_ROOT\n");
	if (N->type == INTERN)
		printf("Type = INTERN\n");
	if (N->type == LEAF)
		printf("Type = LEAF\n");
	printf("nro_entries = %d\n", N->nro_entries);
	printf("E_parent = %p\n", N->E_parent);
	printf("N_parent = %p\n", N->N_parent);
	printf("data : \n");
	if (N->type == LEAF || N->type == LEAF_ROOT)
	{
		for (i=0; i<N->nro_entries; i++)
		{
			printf("\tinfo: [%s]\n", N->data.leaf[i].info);
			printf("\tDparent: %d\n", N->data.leaf[i].Dparent);
			printf("\toid: %d\n\n", N->data.leaf[i].oid);
		}
	}
	else
	{
		for (i=0; i<N->nro_entries; i++)
		{
			printf("\tinfo: [%s]\n", N->data.intern[i].info);
			printf("\trc: %d\n", N->data.intern[i].rc);
			printf("\tDparent: %d\n", N->data.intern[i].Dparent);
			printf("\tchild: %p\n\n", N->data.intern[i].child);
		}
		for (i=0; i<N->nro_entries; i++)
			visit(N->data.intern[i].child);
	}
	printf("--------------------------------\n\n");

	return;
}



